# Operators
OP_SEARCH: str
OP_AND: str
OP_OR: str
OP_PHRASE: str
OP_PROXIMITY: str
OP_PROXIMITY_ORD: str
OP_TERM: str

# Functions
FN_ALL: str
FN_TERM: str
FN_PARSE: str
FN_PHRASE_PREFIX: str
FN_REGEX_PHRASE: str
FN_RANGE_TERM: str
FN_REGEX: str
FN_PROXIMITY: str
FN_PROX_REGEX: str
FN_PROX_ARRAY: str
FN_MORE_LIKE_THIS: str
FN_EMPTY: str
FN_EXISTS: str
FN_FUZZY_TERM: str
FN_PARSE_WITH_FIELD: str
FN_RANGE: str
FN_TERM_SET: str
FN_SCORE: str
FN_SNIPPET: str
FN_SNIPPETS: str
FN_SNIPPET_POSITIONS: str
FN_AGG: str

# Types
PDB_TYPE_BOOST: str
PDB_TYPE_CONST: str
PDB_TYPE_FUZZY: str
PDB_TYPE_SLOP: str
PDB_TYPE_QUERY: str
PDB_TYPE_TOKENIZER_ALIAS: str
PDB_TYPE_TOKENIZER_CHINESE_COMPATIBLE: str
PDB_TYPE_TOKENIZER_JIEBA: str
PDB_TYPE_TOKENIZER_LINDERA: str
PDB_TYPE_TOKENIZER_LITERAL: str
PDB_TYPE_TOKENIZER_LITERAL_NORMALIZED: str
PDB_TYPE_TOKENIZER_NGRAM: str
PDB_TYPE_TOKENIZER_REGEX: str
PDB_TYPE_TOKENIZER_SIMPLE: str
PDB_TYPE_TOKENIZER_SOURCE_CODE: str
PDB_TYPE_TOKENIZER_UNICODE_WORDS: str
PDB_TYPE_TOKENIZER_WHITESPACE: str
